Require Import Utf8.
Require Import List.

Definition monotone (P : nat → Prop) := ∀ n, P (S n) → P n.

Definition IProp := { P : nat → Prop | monotone P }.

Fixpoint IRel (l : list Type) : Type :=
  match l with
  | nil    => IProp
  | A :: l => A → IRel l
  end.

Inductive I_valid_at (n : nat) (P : IProp) : Prop :=
| Valid : proj1_sig P n → I_valid_at n P
.

Definition I_valid (P : IProp) := ∀ n, I_valid_at n P.

Notation "n ⊨ P" := (I_valid_at n P) (at level 98, no associativity).
Notation "⊨ P" := (I_valid P) (at level 98, no associativity).

Lemma I_valid_monotone_S (P : IProp) (n : nat) :
  (S n ⊨ P) → (n ⊨ P).
Proof.
intro H; destruct H; constructor; destruct P; simpl; auto.
Qed.

Lemma I_valid_monotone (P : IProp) (n m : nat) :
  n ≤ m → (m ⊨ P) → (n ⊨ P).
Proof.
induction 1; trivial.
intro; apply IHle; apply I_valid_monotone_S; assumption.
Qed.

(* ========================================================================= *)
(* Embeding Prop *)

Definition I_Prop (P : Prop) : IProp.
exists (λ _, P); unfold monotone; auto.
Defined.

Notation "( P )ᵢ" := (I_Prop P).

Lemma I_Prop_intro (P : Prop) n : P → (n ⊨ (P)ᵢ).
Proof.
intro H; constructor; assumption.
Qed.

Lemma I_Prop_elim (P : Prop) n : (n ⊨ (P)ᵢ) → P.
Proof.
intro H; destruct H; apply H.
Qed.